@@ -11,10 +11,10 @@ from collections import deque, namedtuple
 from tornado.concurrent import Future
 from tornado.ioloop import IOLoop
 
-from distributed.comm.core import Comm, CommClosedError, Connector, Listener
+from distributed.comm.core import BaseListener, Comm, CommClosedError, Connector
 from distributed.comm.registry import Backend, backends
 from distributed.protocol import nested_deserialize
-from distributed.utils import get_ip
+from distributed.utils import get_ip, is_python_shutting_down
 
 logger = logging.getLogger(__name__)
 
@@ -184,7 +184,13 @@ class InProc(Comm):
         self._initialized = True
 
     def _get_finalizer(self):
-        def finalize(write_q=self._write_q, write_loop=self._write_loop, r=repr(self)):
+        r = repr(self)
+
+        def finalize(
+            read_q=self._read_q, write_q=self._write_q, write_loop=self._write_loop, r=r
+        ):
+            if read_q.peek(None) is _EOF or is_python_shutting_down():
+                return
             logger.warning(f"Closing dangling queue in {r}")
             write_loop.add_callback(write_q.put_nowait, _EOF)
 
@@ -198,6 +204,10 @@ class InProc(Comm):
     def peer_address(self) -> str:
         return self._peer_addr
 
+    @property
+    def same_host(self) -> bool:
+        return True
+
     async def read(self, deserializers="ignored"):
         if self._closed:
             raise CommClosedError()
@@ -251,10 +261,11 @@ class InProc(Comm):
             return False
 
 
-class InProcListener(Listener):
+class InProcListener(BaseListener):
     prefix = "inproc"
 
     def __init__(self, address, comm_handler, deserialize=True):
+        super().__init__()
         self.manager = global_manager
         self.address = address or self.manager.new_address()
         self.comm_handler = comm_handler
