@@ -1,6 +1,6 @@
 from __future__ import annotations
 
-import abc
+import asyncio
 import contextlib
 import functools
 import logging
@@ -8,15 +8,23 @@ import os
 import socket
 import subprocess
 import sys
+import tempfile
 import uuid
 import zipfile
 from collections.abc import Awaitable
-from typing import TYPE_CHECKING, Any, ClassVar
+from typing import TYPE_CHECKING, Any, Callable, ClassVar
 
+from dask.typing import Key
 from dask.utils import funcname, tmpfile
 
+from distributed.protocol.pickle import dumps
+
 if TYPE_CHECKING:
-    from distributed.scheduler import Scheduler, TaskStateState  # circular imports
+    # circular imports
+    from distributed.scheduler import Scheduler
+    from distributed.scheduler import TaskStateState as SchedulerTaskStateState
+    from distributed.worker import Worker
+    from distributed.worker_state_machine import TaskStateState as WorkerTaskStateState
 
 logger = logging.getLogger(__name__)
 
@@ -24,7 +32,8 @@ logger = logging.getLogger(__name__)
 class SchedulerPlugin:
     """Interface to extend the Scheduler
 
-    A plugin enables custom hooks to run when specific events occur.  The    scheduler will run the methods of this plugin whenever the corresponding
+    A plugin enables custom hooks to run when specific events occur.
+    The scheduler will run the methods of this plugin whenever the corresponding
     method of the scheduler is run.  This runs user code within the scheduler
     thread that can perform arbitrary operations in synchrony with the scheduler
     itself.
@@ -34,9 +43,14 @@ class SchedulerPlugin:
 
     To implement a plugin:
 
-    1. subclass this class
+    1. inherit from this class
     2. override some of its methods
-    3. add the plugin to the scheduler with ``Scheduler.add_plugin(myplugin)``.
+    3. register the plugin using :meth:`Client.register_plugin<distributed.Client.register_plugin>`.
+
+    The ``idempotent`` attribute is used to control whether or not the plugin should
+    be ignored upon registration if a scheduler plugin with the same name already exists.
+    If ``True``, the plugin is ignored, otherwise the existing plugin is replaced.
+    Defaults to ``False``.
 
     Examples
     --------
@@ -55,6 +69,8 @@ class SchedulerPlugin:
     >>> scheduler.add_plugin(plugin)  # doctest: +SKIP
     """
 
+    idempotent: bool = False
+
     async def start(self, scheduler: Scheduler) -> None:
         """Run when the scheduler starts up
 
@@ -76,11 +92,11 @@ class SchedulerPlugin:
         scheduler: Scheduler,
         *,
         client: str,
-        keys: set[str],
-        tasks: list[str],
-        annotations: dict[str, dict[str, Any]],
-        priority: dict[str, tuple[int | float, ...]],
-        dependencies: dict[str, set],
+        keys: set[Key],
+        tasks: list[Key],
+        annotations: dict[str, dict[Key, Any]],
+        priority: dict[Key, tuple[int | float, ...]],
+        dependencies: dict[Key, set[Key]],
         **kwargs: Any,
     ) -> None:
         """Run when a new graph / tasks enter the scheduler
@@ -119,9 +135,9 @@ class SchedulerPlugin:
 
     def transition(
         self,
-        key: str,
-        start: TaskStateState,
-        finish: TaskStateState,
+        key: Key,
+        start: SchedulerTaskStateState,
+        finish: SchedulerTaskStateState,
         *args: Any,
         stimulus_id: str,
         **kwargs: Any,
@@ -138,13 +154,13 @@ class SchedulerPlugin:
 
         Parameters
         ----------
-        key : string
-        start : string
+        key :
+        start :
             Start state of the transition.
             One of released, waiting, processing, memory, error.
-        finish : string
+        finish :
             Final state of the transition.
-        stimulus_id: string
+        stimulus_id :
             ID of stimulus causing the transition.
         *args, **kwargs :
             More options passed when transitioning
@@ -203,9 +219,16 @@ class WorkerPlugin:
     an event happens, the corresponding method on this class will be called. Note that the
     user code always runs within the Worker's main thread.
 
-    To implement a plugin implement some of the methods of this class and register
-    the plugin to your client in order to have it attached to every existing and
-    future workers with ``Client.register_worker_plugin``.
+    To implement a plugin:
+
+    1. inherit from this class
+    2. override some of its methods
+    3. register the plugin using :meth:`Client.register_plugin<distributed.Client.register_plugin>`.
+
+    The ``idempotent`` attribute is used to control whether or not the plugin should
+    be ignored upon registration if a worker plugin with the same name already exists.
+    If ``True``, the plugin is ignored, otherwise the existing plugin is replaced.
+    Defaults to ``False``.
 
     Examples
     --------
@@ -227,21 +250,29 @@ class WorkerPlugin:
 
     >>> import logging
     >>> plugin = ErrorLogger(logging)
-    >>> client.register_worker_plugin(plugin)  # doctest: +SKIP
+    >>> client.register_plugin(plugin)  # doctest: +SKIP
     """
 
-    def setup(self, worker):
+    idempotent: bool = False
+
+    def setup(self, worker: Worker) -> None | Awaitable[None]:
         """
         Run when the plugin is attached to a worker. This happens when the plugin is registered
         and attached to existing workers, or when a worker is created after the plugin has been
         registered.
         """
 
-    def teardown(self, worker):
+    def teardown(self, worker: Worker) -> None | Awaitable[None]:
         """Run when the worker to which the plugin is attached is closed, or
         when the plugin is removed."""
 
-    def transition(self, key, start, finish, **kwargs):
+    def transition(
+        self,
+        key: Key,
+        start: WorkerTaskStateState,
+        finish: WorkerTaskStateState,
+        **kwargs: Any,
+    ) -> None:
         """
         Throughout the lifecycle of a task (see :doc:`Worker State
         <worker-state>`), Workers are instructed by the scheduler to compute
@@ -257,13 +288,14 @@ class WorkerPlugin:
 
         Parameters
         ----------
-        key : string
-        start : string
+        key :
+        start :
             Start state of the transition.
             One of waiting, ready, executing, long-running, memory, error.
-        finish : string
+        finish :
             Final state of the transition.
-        kwargs : More options passed when transitioning
+        kwargs :
+            More options passed when transitioning
         """
 
 
@@ -275,10 +307,16 @@ class NannyPlugin:
     to run code before the worker is started, or to restart the worker if
     necessary.
 
-    To implement a plugin implement some of the methods of this class and register
-    the plugin to your client in order to have it attached to every existing and
-    future nanny by passing ``nanny=True`` to
-    :meth:`Client.register_worker_plugin<distributed.Client.register_worker_plugin>`.
+    To implement a plugin:
+
+    1. inherit from this class
+    2. override some of its methods
+    3. register the plugin using :meth:`Client.register_plugin<distributed.Client.register_plugin>`.
+
+    The ``idempotent`` attribute is used to control whether or not the plugin should
+    be ignored upon registration if a nanny plugin with the same name already exists.
+    If ``True``, the plugin is ignored, otherwise the existing plugin is replaced.
+    Defaults to ``False``.
 
     The ``restart`` attribute is used to control whether or not a running ``Worker``
     needs to be restarted when registering the plugin.
@@ -289,7 +327,8 @@ class NannyPlugin:
     SchedulerPlugin
     """
 
-    restart = False
+    idempotent: bool = False
+    restart: bool = False
 
     def setup(self, nanny):
         """
@@ -330,25 +369,25 @@ class SchedulerUploadFile(SchedulerPlugin):
         await scheduler.upload_file(self.filename, self.data, load=self.load)
 
 
-class PackageInstall(WorkerPlugin, abc.ABC):
-    """Abstract parent class for a worker plugin to install a set of packages
+class InstallPlugin(SchedulerPlugin):
+    """Scheduler plugin to install software on the cluster
 
-    This accepts a set of packages to install on all workers.
-    You can also optionally ask for the worker to restart itself after
-    performing this installation.
+    This accepts an function that installs software on the scheduler and
+    all workers. You can also optionally ask for the worker to restart
+    after performing this installation.
 
     .. note::
 
        This will increase the time it takes to start up
        each worker. If possible, we recommend including the
-       libraries in the worker environment or image. This is
+       software in the worker environment or image. This is
        primarily intended for experimentation and debugging.
 
     Parameters
     ----------
-    packages
-        A list of packages (with optional versions) to install
-    restart
+    install_fn
+        Callable used to install the software; must be idempotent.
+    restart_workers
         Whether or not to restart the worker after installing the packages
         Only functions if the worker has an attached nanny process
 
@@ -358,20 +397,138 @@ class PackageInstall(WorkerPlugin, abc.ABC):
     PipInstall
     """
 
-    INSTALLER: ClassVar[str]
+    idempotent = True
+    _lock: ClassVar[asyncio.Lock | None] = None
 
+    _install_fn: Callable[[], None]
     name: str
-    packages: list[str]
-    restart: bool
+    restart_workers: bool
+    _scheduler: Scheduler
 
     def __init__(
         self,
-        packages: list[str],
-        restart: bool,
+        install_fn: Callable[[], None],
+        restart_workers: bool,
     ):
-        self.packages = packages
-        self.restart = restart
-        self.name = f"{self.INSTALLER}-install-{uuid.uuid4()}"
+        self._install_fn = install_fn
+        self.restart_workers = restart_workers
+        self.name = f"{self.__class__.__name__}-{uuid.uuid4()}"
+
+    async def start(self, scheduler: Scheduler) -> None:
+        self._scheduler = scheduler
+
+        if InstallPlugin._lock is None:
+            InstallPlugin._lock = asyncio.Lock()
+
+        async with InstallPlugin._lock:
+            self._install_fn()
+
+            if self.restart_workers:
+                nanny_plugin = _InstallNannyPlugin(self._install_fn, self.name)
+                await scheduler.register_nanny_plugin(
+                    comm=None,
+                    plugin=dumps(nanny_plugin),
+                    name=self.name,
+                    idempotent=True,
+                )
+            else:
+                worker_plugin = _InstallWorkerPlugin(self._install_fn, self.name)
+                await scheduler.register_worker_plugin(
+                    comm=None,
+                    plugin=dumps(worker_plugin),
+                    name=self.name,
+                    idempotent=True,
+                )
+
+    async def close(self) -> None:
+        assert InstallPlugin._lock is not None
+
+        async with InstallPlugin._lock:
+            if self.restart_workers:
+                await self._scheduler.unregister_nanny_plugin(comm=None, name=self.name)
+            else:
+                await self._scheduler.unregister_worker_plugin(
+                    comm=None, name=self.name
+                )
+
+
+class _InstallNannyPlugin(NannyPlugin):
+    """Nanny plugin to install software on the nanny and then restart the workers
+
+    This accepts an installer which will installs software on all nannies.
+
+    .. note::
+
+       This will increase the time it takes to start up
+       the cluster. If possible, we recommend including the
+       software in the cluster environment or image. This is
+       primarily intended for experimentation and debugging.
+
+    Parameters
+    ----------
+    install_fn
+        Function that installs software
+    name:
+        Name of the plugin
+
+    See Also
+    --------
+    InstallPlugin
+    """
+
+    _install_fn: Callable[[], None]
+    name: str
+    restart = True
+
+    def __init__(self, _install_fn: Callable[[], None], name: str):
+        self._install_fn = _install_fn
+        self.name = name
+
+    async def setup(self, nanny):
+        from distributed.semaphore import Semaphore
+
+        async with (
+            await Semaphore(
+                max_leases=1,
+                name=socket.gethostname(),
+                register=True,
+                scheduler_rpc=nanny.scheduler,
+                loop=nanny.loop,
+            )
+        ):
+            self._install_fn()
+
+
+class _InstallWorkerPlugin(WorkerPlugin):
+    """Worker plugin to install software on the worker
+
+    This accepts an installer which will installs software on all workers.
+
+    .. note::
+
+       This will increase the time it takes to start up
+       the cluster. If possible, we recommend including the
+       software in the cluster environment or image. This is
+       primarily intended for experimentation and debugging.
+
+    Parameters
+    ----------
+    install_fn
+        Callable that installs software
+    name:
+        Name of the plugin
+
+    See Also
+    --------
+    InstallPlugin
+    """
+
+    _install_fn: Callable[[], None]
+    name: str
+
+    def __init__(self, install_fn: Callable[[], None], name: str):
+        self._install_fn = install_fn
+        self.name = name
 
     async def setup(self, worker):
         from distributed.semaphore import Semaphore
@@ -385,78 +542,23 @@ class PackageInstall(WorkerPlugin, abc.ABC):
                 loop=worker.loop,
             )
         ):
-            if not await self._is_installed(worker):
-                logger.info(
-                    "%s installing the following packages: %s",
-                    self.INSTALLER,
-                    self.packages,
-                )
-                await self._set_installed(worker)
-                self.install()
-            else:
-                logger.info(
-                    "The following packages have already been installed: %s",
-                    self.packages,
-                )
-
-            if self.restart and worker.nanny and not await self._is_restarted(worker):
-                logger.info("Restarting worker to refresh interpreter.")
-                await self._set_restarted(worker)
-                worker.loop.add_callback(
-                    worker.close_gracefully, restart=True, reason=f"{self.name}-setup"
-                )
-
-    @abc.abstractmethod
-    def install(self) -> None:
-        """Install the requested packages"""
-
-    async def _is_installed(self, worker):
-        return await worker.client.get_metadata(
-            self._compose_installed_key(), default=False
-        )
-
-    async def _set_installed(self, worker):
-        await worker.client.set_metadata(
-            self._compose_installed_key(),
-            True,
-        )
-
-    def _compose_installed_key(self):
-        return [
-            self.name,
-            "installed",
-            socket.gethostname(),
-        ]
-
-    async def _is_restarted(self, worker):
-        return await worker.client.get_metadata(
-            self._compose_restarted_key(worker),
-            default=False,
-        )
-
-    async def _set_restarted(self, worker):
-        await worker.client.set_metadata(
-            self._compose_restarted_key(worker),
-            True,
-        )
-
-    def _compose_restarted_key(self, worker):
-        return [self.name, "restarted", worker.nanny]
+            self._install_fn()
 
 
-class CondaInstall(PackageInstall):
-    """A Worker Plugin to conda install a set of packages
+class CondaInstall(InstallPlugin):
+    """A plugin to conda install a set of packages
 
-    This accepts a set of packages to install on all workers as well as
-    options to use when installing.
-    You can also optionally ask for the worker to restart itself after
+    This accepts a set of packages to install on the scheduler and all
+    workers as well as options to use when installing.
+
+    You can also optionally ask for the workers to restart after
     performing this installation.
 
     .. note::
 
        This will increase the time it takes to start up
-       each worker. If possible, we recommend including the
-       libraries in the worker environment or image. This is
+       the cluster. If possible, we recommend including the
+       libraries in the cluster environment or image. This is
        primarily intended for experimentation and debugging.
 
     Parameters
@@ -465,37 +567,49 @@ class CondaInstall(PackageInstall):
         A list of packages (with optional versions) to install using conda
     conda_options
         Additional options to pass to conda
-    restart
+    restart_workers
         Whether or not to restart the worker after installing the packages
-        Only functions if the worker has an attached nanny process
+        Only functions if the workers have an attached nanny process
 
     Examples
     --------
     >>> from dask.distributed import CondaInstall
     >>> plugin = CondaInstall(packages=["scikit-learn"], conda_options=["--update-deps"])
 
-    >>> client.register_worker_plugin(plugin)
+    >>> client.register_plugin(plugin)
 
     See Also
     --------
-    PackageInstall
+    InstallPlugin
     PipInstall
     """
 
-    INSTALLER = "conda"
-
-    conda_options: list[str]
-
     def __init__(
         self,
         packages: list[str],
         conda_options: list[str] | None = None,
-        restart: bool = False,
+        restart_workers: bool = False,
     ):
-        super().__init__(packages, restart=restart)
+        installer = _CondaInstaller(packages, conda_options)
+        super().__init__(installer, restart_workers=restart_workers)
+
+
+class _CondaInstaller:
+    INSTALLER = "conda"
+
+    packages: list[str]
+    conda_options: list[str]
+
+    def __init__(self, packages: list[str], conda_options: list[str] | None = None):
+        self.packages = packages
         self.conda_options = conda_options or []
 
-    def install(self) -> None:
+    def __call__(self) -> None:
+        logger.info(
+            "%s installing the following packages: %s",
+            self.INSTALLER,
+            self.packages,
+        )
         try:
             from conda.cli.python_api import Commands, run_command
         except ModuleNotFoundError as e:  # pragma: nocover
@@ -520,8 +634,8 @@ class CondaInstall(PackageInstall):
             raise RuntimeError(msg)
 
 
-class PipInstall(PackageInstall):
-    """A Worker Plugin to pip install a set of packages
+class PipInstall(InstallPlugin):
+    """A plugin to pip install a set of packages
 
     This accepts a set of packages to install on all workers as well as
     options to use when installing.
@@ -538,51 +652,86 @@ class PipInstall(PackageInstall):
     Parameters
     ----------
     packages
-        A list of packages (with optional versions) to install using pip
+        A list of packages to install using pip.
+        Packages should follow the structure defined for
+        `requirement files <https://pip.pypa.io/en/stable/reference/requirements-file-format/#structure>`_.
+        Packages also may include
+        `environment variables <https://pip.pypa.io/en/stable/reference/requirements-file-format/#using-environment-variables>`_.
     pip_options
         Additional options to pass to pip
-    restart
-        Whether or not to restart the worker after installing the packages
-        Only functions if the worker has an attached nanny process
+    restart_workers
+        Whether or not to restart the worker after installing the packages;
+        only functions if the worker has an attached nanny process.
 
     Examples
     --------
     >>> from dask.distributed import PipInstall
     >>> plugin = PipInstall(packages=["scikit-learn"], pip_options=["--upgrade"])
+    >>> client.register_plugin(plugin)
 
-    >>> client.register_worker_plugin(plugin)
+    Install package from a private repository using a ``TOKEN`` environment variable.
+
+    >>> from dask.distributed import PipInstall
+    >>> plugin = PipInstall(packages=["private_package@git+https://${TOKEN}@github.com/dask/private_package.git])
+    >>> client.register_plugin(plugin)
 
     See Also
     --------
-    PackageInstall
+    InstallPlugin
     CondaInstall
     """
 
-    INSTALLER = "pip"
-
-    pip_options: list[str]
-
     def __init__(
         self,
         packages: list[str],
         pip_options: list[str] | None = None,
-        restart: bool = False,
+        restart_workers: bool = False,
     ):
-        super().__init__(packages, restart=restart)
+        installer = _PipInstaller(packages, pip_options)
+        super().__init__(installer, restart_workers=restart_workers)
+
+
+class _PipInstaller:
+    INSTALLER = "pip"
+
+    packages: list[str]
+    pip_options: list[str]
+
+    def __init__(self, packages: list[str], pip_options: list[str] | None = None):
+        self.packages = packages
         self.pip_options = pip_options or []
 
-    def install(self) -> None:
-        proc = subprocess.Popen(
-            [sys.executable, "-m", "pip", "install"] + self.pip_options + self.packages,
-            stdout=subprocess.PIPE,
-            stderr=subprocess.PIPE,
+    def __call__(self) -> None:
+        logger.info(
+            "%s installing the following packages: %s",
+            self.INSTALLER,
+            self.packages,
         )
-        _, stderr = proc.communicate()
-        returncode = proc.wait()
-        if returncode != 0:
-            msg = f"pip install failed with '{stderr.decode().strip()}'"
-            logger.error(msg)
-            raise RuntimeError(msg)
+        # Use a requirements file under the hood to support
+        # environment variables
+        # See https://pip.pypa.io/en/stable/reference/requirements-file-format/#using-environment-variables
+        with tempfile.NamedTemporaryFile(mode="w+") as f:
+            f.write("\n".join(self.packages))
+            f.flush()
+            proc = subprocess.Popen(
+                [
+                    sys.executable,
+                    "-m",
+                    "pip",
+                    "install",
+                    *self.pip_options,
+                    "-r",
+                    f.name,
+                ],
+                stdout=subprocess.PIPE,
+                stderr=subprocess.PIPE,
+            )
+            _, stderr = proc.communicate()
+            returncode = proc.wait()
+            if returncode != 0:
+                msg = f"pip install failed with '{stderr.decode().strip()}'"
+                logger.error(msg)
+                raise RuntimeError(msg)
 
 
 # Adapted from https://github.com/dask/distributed/issues/3560#issuecomment-596138522
@@ -598,7 +747,7 @@ class UploadFile(WorkerPlugin):
     --------
     >>> from distributed.diagnostics.plugin import UploadFile
 
-    >>> client.register_worker_plugin(UploadFile("/path/to/file.py"))  # doctest: +SKIP
+    >>> client.register_plugin(UploadFile("/path/to/file.py"))  # doctest: +SKIP
     """
 
     name = "upload_file"
@@ -722,7 +871,7 @@ class UploadDirectory(NannyPlugin):
     Examples
     --------
     >>> from distributed.diagnostics.plugin import UploadDirectory
-    >>> client.register_worker_plugin(UploadDirectory("/path/to/directory"), nanny=True)  # doctest: +SKIP
+    >>> client.register_plugin(UploadDirectory("/path/to/directory"), nanny=True)  # doctest: +SKIP
     """
 
     def __init__(
@@ -859,7 +1008,7 @@ class ForwardOutput(WorkerPlugin):
     >>> from dask.distributed import ForwardOutput
     >>> plugin = ForwardOutput()
 
-    >>> client.register_worker_plugin(plugin)
+    >>> client.register_plugin(plugin)
     """
 
     def setup(self, worker):
