@@ -1,28 +1,36 @@
 from __future__ import annotations
 
+import codecs
 import importlib
 import traceback
 from array import array
 from enum import Enum
 from functools import partial
+from pickle import PickleBuffer
 from types import ModuleType
+from typing import Any, Generic, Literal, TypeVar
 
 import msgpack
 
 import dask
 from dask.base import normalize_token
+from dask.sizeof import sizeof
 from dask.utils import typename
 
+from distributed.metrics import context_meter
 from distributed.protocol import pickle
 from distributed.protocol.compression import decompress, maybe_compress
 from distributed.protocol.utils import (
     frame_split_size,
+    host_array_from_buffers,
     merge_memoryviews,
     msgpack_opts,
     pack_frames_prelude,
     unpack_frames,
 )
-from distributed.utils import ensure_bytes, has_keyword
+from distributed.utils import ensure_memoryview, has_keyword
+
+T = TypeVar("T")
 
 dask_serialize = dask.utils.Dispatch("dask_serialize")
 dask_deserialize = dask.utils.Dispatch("dask_deserialize")
@@ -45,7 +53,7 @@ def dask_dumps(x, context=None):
     header = {
         "sub-header": sub_header,
         "type": type_name,
-        "type-serialized": pickle.dumps(type(x), protocol=4),
+        "type-serialized": pickle.dumps(type(x)),
         "serializer": "dask",
     }
     return header, frames
@@ -59,7 +67,13 @@ def dask_loads(header, frames):
 
 def pickle_dumps(x, context=None):
     frames = [None]
-    buffer_callback = lambda f: frames.append(memoryview(f))
+    writeable = []
+
+    def buffer_callback(f):
+        f = memoryview(f)
+        frames.append(f)
+        writeable.append(not f.readonly)
+
     frames[0] = pickle.dumps(
         x,
         buffer_callback=buffer_callback,
@@ -67,33 +81,33 @@ def pickle_dumps(x, context=None):
     )
     header = {
         "serializer": "pickle",
-        "writeable": tuple(not f.readonly for f in frames[1:]),
+        "writeable": tuple(writeable),
     }
+
     return header, frames
 
 
-def pickle_loads(header, frames):
-    x, buffers = frames[0], frames[1:]
+def pickle_loads(
+    header: dict[str, Any], frames: list[bytes | bytearray | memoryview | PickleBuffer]
+) -> Any:
+    pik, buffers = frames[0], frames[1:]
 
-    writeable = header.get("writeable")
-    if not writeable:
-        writeable = len(buffers) * (None,)
+    def ensure_writeable_flag(mv: memoryview, w: bool) -> memoryview:
+        if w and mv.readonly:
+            # Can't avoid a deep copy
+            return memoryview(bytearray(mv))
+        elif not w and not mv.readonly:
+            # Zero copy - this is just a flag
+            return mv.toreadonly()
+        else:
+            return mv
 
-    new = []
-    memoryviews = map(memoryview, buffers)
-    for w, mv in zip(writeable, memoryviews):
-        if w == mv.readonly:
-            if mv.readonly:
-                mv = memoryview(bytearray(mv))
-            else:
-                mv = memoryview(bytes(mv))
-            if mv.nbytes > 0:
-                mv = mv.cast(mv.format, mv.shape)
-            else:
-                mv = mv.cast(mv.format)
-        new.append(mv)
+    buffers = [
+        ensure_writeable_flag(ensure_memoryview(mv), w)
+        for mv, w in zip(buffers, header["writeable"])
+    ]
 
-    return pickle.loads(x, buffers=new)
+    return pickle.loads(pik, buffers=buffers)
 
 
 def import_allowed_module(name):
@@ -176,7 +190,7 @@ def msgpack_loads(header, frames):
 
 
 def serialization_error_loads(header, frames):
-    msg = "\n".join([ensure_bytes(frame).decode("utf8") for frame in frames])
+    msg = "\n".join([codecs.decode(frame, "utf8") for frame in frames])
     raise TypeError(msg)
 
 
@@ -207,9 +221,13 @@ def check_dask_serializable(x):
     return False
 
 
-def serialize(
-    x, serializers=None, on_error="message", context=None, iterate_collection=None
-):
+def serialize(  # type: ignore[no-untyped-def]
+    x: object,
+    serializers=None,
+    on_error: Literal["message" | "raise"] = "message",
+    context=None,
+    iterate_collection: bool | None = None,
+) -> tuple[dict[str, Any], list[bytes | memoryview]]:
     r"""
     Convert object to a header and list of bytestrings
 
@@ -254,7 +272,7 @@ def serialize(
     if serializers is None:
         serializers = ("dask", "pickle")  # TODO: get from configuration
 
-    # Handle obects that are marked as `Serialize`, or that are
+    # Handle objects that are marked as `Serialize`, or that are
     # already `Serialized` objects (don't want to serialize them twice)
     if isinstance(x, Serialized):
         return x.header, x.frames
@@ -267,6 +285,9 @@ def serialize(
             iterate_collection=True,
         )
 
+    # Note: don't use isinstance(), as it would match subclasses
+    # (e.g. namedtuple, defaultdict) which however would revert to the base class on a
+    # round-trip through msgpack
     if iterate_collection is None and type(x) in (list, set, tuple, dict):
         if type(x) is list and "msgpack" in serializers:
             # Note: "msgpack" will always convert lists to tuples
@@ -305,6 +326,7 @@ def serialize(
                 _header["key"] = k
                 headers_frames.append((_header, _frames))
         else:
+            assert isinstance(x, (list, set, tuple))
             headers_frames = [
                 serialize(
                     obj, serializers=serializers, on_error=on_error, context=context
@@ -314,16 +336,15 @@ def serialize(
 
         frames = []
         lengths = []
-        compressions = []
+        compressions: list[str | None] = []
         for _header, _frames in headers_frames:
             frames.extend(_frames)
             length = len(_frames)
             lengths.append(length)
             compressions.extend(_header.get("compression") or [None] * len(_frames))
 
-        headers = [obj[0] for obj in headers_frames]
         headers = {
-            "sub-headers": headers,
+            "sub-headers": [obj[0] for obj in headers_frames],
             "is-collection": True,
             "frame-lengths": lengths,
             "type-serialized": type(x).__name__,
@@ -333,6 +354,7 @@ def serialize(
         return headers, frames
 
     tb = ""
+    exc = None
 
     for name in serializers:
         dumps, _, wants_context = families[name]
@@ -342,21 +364,30 @@ def serialize(
             return header, frames
         except NotImplementedError:
             continue
-        except Exception:
+        except Exception as e:
+            exc = e
             tb = traceback.format_exc()
             break
-
-    msg = "Could not serialize object of type %s." % type(x).__name__
+    type_x = type(x)
+    if isinstance(x, (ToPickle, Serialize)):
+        type_x = type(x.data)
+    msg = f"Could not serialize object of type {type_x.__name__}"
     if on_error == "message":
-        frames = [msg]
+        txt_frames = [msg]
         if tb:
-            frames.append(tb[:100000])
+            txt_frames.append(tb[:100000])
 
-        frames = [frame.encode() for frame in frames]
+        frames = [frame.encode() for frame in txt_frames]
 
         return {"serializer": "error"}, frames
     elif on_error == "raise":
-        raise TypeError(msg, str(x)[:10000])
+        try:
+            str_x = str(x)[:10000]
+        except Exception:
+            raise TypeError(msg) from exc
+        raise TypeError(msg, str_x) from exc
+    else:  # pragma: nocover
+        raise ValueError(f"{on_error=}; expected 'message' or 'raise'")
 
 
 def deserialize(header, frames, deserializers=None):
@@ -417,10 +448,11 @@ def deserialize(header, frames, deserializers=None):
     return loads(header, frames)
 
 
+@context_meter.meter("serialize")
 def serialize_and_split(
     x, serializers=None, on_error="message", context=None, size=None
 ):
-    """Serialize and split compressable frames
+    """Serialize and split compressible frames
 
     This function is a drop-in replacement of `serialize()` that calls `serialize()`
     followed by `frame_split_size()` on frames that should be compressed.
@@ -453,7 +485,7 @@ def serialize_and_split(
             out_compression.append(compression)
     assert len(out_compression) == len(out_frames)
 
-    # Notice, in order to match msgpack's implicit convertion to tuples,
+    # Notice, in order to match msgpack's implicit conversion to tuples,
     # we convert to tuples here as well.
     header["split-num-sub-frames"] = tuple(num_sub_frames)
     header["split-offsets"] = tuple(offsets)
@@ -461,6 +493,7 @@ def serialize_and_split(
     return header, out_frames
 
 
+@context_meter.meter("deserialize")
 def merge_and_deserialize(header, frames, deserializers=None):
     """Merge and deserialize frames
 
@@ -481,7 +514,7 @@ def merge_and_deserialize(header, frames, deserializers=None):
             try:
                 merged = merge_memoryviews(subframes)
             except (ValueError, TypeError):
-                merged = bytearray().join(subframes)
+                merged = host_array_from_buffers(subframes)
 
             merged_frames.append(merged)
 
@@ -522,8 +555,7 @@ to_serialize = Serialize
 
 
 class Serialized:
-    """
-    An object that is already serialized into header and frames
+    """An object that is already serialized into header and frames
 
     Normal serialization operations pass these objects through.  This is
     typically used within the scheduler which accepts messages that contain
@@ -545,6 +577,49 @@ class Serialized:
         return not (self == other)
 
 
+class ToPickle(Generic[T]):
+    """Mark an object that should be pickled
+
+    Both the scheduler and workers with automatically unpickle this
+    object on arrival.
+    """
+
+    data: T
+
+    def __init__(self, data: T):
+        self.data = data
+
+    def __repr__(self) -> str:
+        return "<ToPickle: %s>" % str(self.data)
+
+    def __eq__(self, other: object) -> bool:
+        return isinstance(other, type(self)) and other.data == self.data
+
+    def __hash__(self) -> int:
+        return hash(self.data)
+
+
+class Pickled:
+    """An object that is already pickled into header and frames
+
+    Normal pickled objects are unpickled by the scheduler.
+    """
+
+    def __init__(self, header, frames):
+        self.header = header
+        self.frames = frames
+
+    def __eq__(self, other):
+        return (
+            isinstance(other, type(self))
+            and other.header == self.header
+            and other.frames == self.frames
+        )
+
+    def __ne__(self, other):
+        return not (self == other)
+
+
 def nested_deserialize(x):
     """
     Replace all Serialize and Serialized values nested in *x*
@@ -583,13 +658,26 @@ def nested_deserialize(x):
     return replace_inner(x)
 
 
-def serialize_bytelist(x, **kwargs):
+@sizeof.register(ToPickle)
+@sizeof.register(Serialize)
+def sizeof_serialize(obj):
+    return sizeof(obj.data)
+
+
+@sizeof.register(Pickled)
+@sizeof.register(Serialized)
+def sizeof_serialized(obj):
+    return sizeof(obj.header) + sizeof(obj.frames)
+
+
+def serialize_bytelist(
+    x: object, compression: str | None | Literal[False] = "auto", **kwargs: Any
+) -> list[bytes | bytearray | memoryview]:
     header, frames = serialize_and_split(x, **kwargs)
     if frames:
-        compression, frames = zip(*map(maybe_compress, frames))
-    else:
-        compression = []
-    header["compression"] = compression
+        header["compression"], frames = zip(
+            *(maybe_compress(frame, compression=compression) for frame in frames)
+        )
     header["count"] = len(frames)
 
     header = msgpack.dumps(header, use_bin_type=True)
@@ -598,20 +686,21 @@ def serialize_bytelist(x, **kwargs):
     return frames2
 
 
-def serialize_bytes(x, **kwargs):
+def serialize_bytes(x: object, **kwargs: Any) -> bytes:
     L = serialize_bytelist(x, **kwargs)
     return b"".join(L)
 
 
-def deserialize_bytes(b):
+def deserialize_bytes(b: bytes | bytearray | memoryview) -> Any:
+    """Deserialize the output of :func:`serialize_bytes`"""
     frames = unpack_frames(b)
-    header, frames = frames[0], frames[1:]
-    if header:
-        header = msgpack.loads(header, raw=False, use_list=False)
+    bin_header, frames = frames[0], frames[1:]
+    if bin_header:
+        header = msgpack.loads(bin_header, raw=False, use_list=False)
     else:
         header = {}
-    frames = decompress(header, frames)
-    return merge_and_deserialize(header, frames)
+    frames2 = decompress(header, frames)
+    return merge_and_deserialize(header, frames2)
 
 
 ################################
@@ -693,7 +782,7 @@ def _deserialize_bytes(header, frames):
     if len(frames) == 1 and isinstance(frames[0], bytes):
         return frames[0]
     else:
-        return bytes().join(frames)
+        return b"".join(frames)
 
 
 @dask_deserialize.register(bytearray)
@@ -714,12 +803,11 @@ def _serialize_array(obj):
 @dask_deserialize.register(array)
 def _deserialize_array(header, frames):
     a = array(header["typecode"])
-    for f in map(memoryview, frames):
-        try:
-            f = f.cast("B")
-        except TypeError:
-            f = f.tobytes()
-        a.frombytes(f)
+    nframes = len(frames)
+    if nframes == 1:
+        a.frombytes(ensure_memoryview(frames[0]))
+    elif nframes > 1:
+        a.frombytes(b"".join(map(ensure_memoryview, frames)))
     return a
 
 
@@ -727,6 +815,8 @@ def _deserialize_array(header, frames):
 def _serialize_memoryview(obj):
     if obj.format == "O":
         raise ValueError("Cannot serialize `memoryview` containing Python objects")
+    if not obj and obj.ndim > 1:
+        raise ValueError("Cannot serialize empty non-1-D `memoryview`")
     header = {"format": obj.format, "shape": obj.shape}
     frames = [obj]
     return header, frames
@@ -735,10 +825,17 @@ def _serialize_memoryview(obj):
 @dask_deserialize.register(memoryview)
 def _deserialize_memoryview(header, frames):
     if len(frames) == 1:
-        out = memoryview(frames[0]).cast("B")
+        out = ensure_memoryview(frames[0])
     else:
         out = memoryview(b"".join(frames))
-    out = out.cast(header["format"], header["shape"])
+
+    # handle empty `memoryview`s
+    if out:
+        out = out.cast(header["format"], header["shape"])
+    else:
+        out = out.cast(header["format"])
+        assert out.shape == header["shape"]
+
     return out
 
 
@@ -753,16 +850,38 @@ def _is_msgpack_serializable(v):
         v is None
         or typ is str
         or typ is bool
+        or typ is bytes
         or typ is int
         or typ is float
         or isinstance(v, dict)
         and all(map(_is_msgpack_serializable, v.values()))
-        and all(typ is str for x in v.keys())
+        and all(type(x) is str for x in v.keys())
         or isinstance(v, (list, tuple))
         and all(map(_is_msgpack_serializable, v))
     )
 
 
+def _is_dumpable(v):
+    typ = type(v)
+    return (
+        v is None
+        or typ is str
+        or typ is bool
+        or typ is bytes
+        or typ is int
+        or typ is float
+        or typ is Pickled
+        or typ is Serialize
+        or typ is Serialized
+        or typ is ToPickle
+        or isinstance(v, dict)
+        and all(map(_is_dumpable, v.values()))
+        and all(type(x) is str for x in v.keys())
+        or isinstance(v, (list, tuple))
+        and all(map(_is_dumpable, v))
+    )
+
+
 class ObjectDictSerializer:
     def __init__(self, serializer):
         self.serializer = serializer
@@ -770,7 +889,7 @@ class ObjectDictSerializer:
     def serialize(self, est):
         header = {
             "serializer": self.serializer,
-            "type-serialized": pickle.dumps(type(est), protocol=4),
+            "type-serialized": pickle.dumps(type(est)),
             "simple": {},
             "complex": {},
         }
