@@ -2,7 +2,8 @@ from __future__ import annotations
 
 import ctypes
 import struct
-from collections.abc import Sequence
+from collections.abc import Collection, Iterable, Sequence
+from typing import Literal, overload
 
 import dask
 
@@ -11,13 +12,40 @@ from distributed.utils import nbytes
 BIG_BYTES_SHARD_SIZE = dask.utils.parse_bytes(dask.config.get("distributed.comm.shard"))
 
 
-msgpack_opts = {
-    ("max_%s_len" % x): 2**31 - 1 for x in ["str", "bin", "array", "map", "ext"]
-}
+msgpack_opts = {}
 msgpack_opts["strict_map_key"] = False
 msgpack_opts["raw"] = False
 
 
+# Find the function, `host_array()`, to use when allocating new host arrays
+try:
+    # Use NumPy, when available, to avoid memory initialization cost.
+    # A `bytearray` is zero-initialized using `calloc`, which we don't need.
+    # `np.empty` both skips the zero-initialization, and
+    # uses hugepages when available ( https://github.com/numpy/numpy/pull/14216 ).
+    import numpy
+
+    def host_array(n: int) -> memoryview:
+        return numpy.empty((n,), dtype="u1").data
+
+except ImportError:
+
+    def host_array(n: int) -> memoryview:
+        return memoryview(bytearray(n))
+
+
+def host_array_from_buffers(
+    buffers: Iterable[bytes | bytearray | memoryview],
+) -> memoryview:
+    mvs = [memoryview(buf) for buf in buffers]
+    out = host_array(sum(mv.nbytes for mv in mvs))
+    offset = 0
+    for mv in mvs:
+        out[offset : offset + mv.nbytes] = mv.cast("B")
+        offset += mv.nbytes
+    return out
+
+
 def frame_split_size(
     frame: bytes | memoryview, n: int = BIG_BYTES_SHARD_SIZE
 ) -> list[memoryview]:
@@ -43,13 +71,13 @@ def frame_split_size(
     return [frame[i : i + items_per_shard] for i in range(0, nitems, items_per_shard)]
 
 
-def pack_frames_prelude(frames):
+def pack_frames_prelude(frames: Collection[bytes | bytearray | memoryview]) -> bytes:
     nframes = len(frames)
     nbytes_frames = map(nbytes, frames)
     return struct.pack(f"Q{nframes}Q", nframes, *nbytes_frames)
 
 
-def pack_frames(frames):
+def pack_frames(frames: Collection[bytes | bytearray | memoryview]) -> bytes:
     """Pack frames into a byte-like object
 
     This prepends length information to the front of the bytes-like object
@@ -61,12 +89,44 @@ def pack_frames(frames):
     return b"".join([pack_frames_prelude(frames), *frames])
 
 
-def unpack_frames(b):
+@overload
+def unpack_frames(
+    b: bytes | bytearray | memoryview,
+    *,
+    remainder: bool = False,
+    partial: Literal[False] = False,
+) -> list[memoryview]:
+    ...
+
+
+@overload
+def unpack_frames(
+    b: bytes | bytearray | memoryview,
+    *,
+    remainder: bool = False,
+    partial: Literal[True],
+) -> tuple[list[memoryview], list[int]]:
+    ...
+
+
+def unpack_frames(b, *, remainder=False, partial=False):
     """Unpack bytes into a sequence of frames
 
     This assumes that length information is at the front of the bytestring,
     as performed by pack_frames
 
+    Parameters
+    ----------
+    b:
+        packed frames, as returned by :func:`pack_frames`
+    remainder:
+        If True, return one extra frame at the end which is the continuation of a
+        stream created by concatenating multiple calls to :func:`pack_frames`.
+        This last frame will be empty at the end of the stream.
+    partial:
+        If True, allow for b to contain less frames than what the preamble indicates;
+        return a tuple of ([frames so far], [lengths of missing frames])
+
     See Also
     --------
     pack_frames
@@ -81,12 +141,26 @@ def unpack_frames(b):
 
     frames = []
     start = fmt_size * (1 + n_frames)
+    nb = b.nbytes
+    end = 0
+    missing_lengths = []
     for length in lengths:
+        if partial and start == nb:
+            missing_lengths.extend(lengths[len(frames) :])
+            break
+
         end = start + length
         frames.append(b[start:end])
         start = end
+    assert end <= nb
 
-    return frames
+    if remainder:
+        frames.append(b[start:])
+
+    if partial:
+        return frames, missing_lengths
+    else:
+        return frames
 
 
 def merge_memoryviews(mvs: Sequence[memoryview]) -> memoryview:
