@@ -1,26 +1,27 @@
 from __future__ import annotations
 
-import abc
 import contextlib
 import itertools
 import logging
 from collections import defaultdict
-from collections.abc import Callable, Iterable, Sequence
-from dataclasses import dataclass, field
-from functools import partial
-from itertools import product
-from typing import TYPE_CHECKING, Any, ClassVar
+from typing import TYPE_CHECKING, Any
+
+from dask.typing import Key
 
 from distributed.diagnostics.plugin import SchedulerPlugin
+from distributed.metrics import time
 from distributed.protocol.pickle import dumps
-from distributed.shuffle._rechunk import ChunkedAxes, NDIndex
-from distributed.shuffle._shuffle import (
+from distributed.protocol.serialize import ToPickle
+from distributed.shuffle._core import (
+    SchedulerShuffleState,
     ShuffleId,
-    ShuffleType,
+    ShuffleRunSpec,
+    ShuffleSpec,
     barrier_key,
     id_from_key,
 )
 from distributed.shuffle._worker_plugin import ShuffleWorkerPlugin
+from distributed.utils import log_errors
 
 if TYPE_CHECKING:
     from distributed.scheduler import (
@@ -34,63 +35,6 @@ if TYPE_CHECKING:
 logger = logging.getLogger(__name__)
 
 
-@dataclass(eq=False)
-class ShuffleState(abc.ABC):
-    _run_id_iterator: ClassVar[itertools.count] = itertools.count(1)
-
-    id: ShuffleId
-    run_id: int
-    output_workers: set[str]
-    participating_workers: set[str]
-    _archived_by: str | None = field(default=None, init=False)
-
-    @abc.abstractmethod
-    def to_msg(self) -> dict[str, Any]:
-        """Transform the shuffle state into a JSON-serializable message"""
-
-    def __str__(self) -> str:
-        return f"{self.__class__.__name__}<{self.id}[{self.run_id}]>"
-
-    def __hash__(self) -> int:
-        return hash(self.run_id)
-
-
-@dataclass(eq=False)
-class DataFrameShuffleState(ShuffleState):
-    type: ClassVar[ShuffleType] = ShuffleType.DATAFRAME
-    worker_for: dict[int, str]
-    column: str
-
-    def to_msg(self) -> dict[str, Any]:
-        return {
-            "status": "OK",
-            "type": DataFrameShuffleState.type,
-            "run_id": self.run_id,
-            "worker_for": self.worker_for,
-            "column": self.column,
-            "output_workers": self.output_workers,
-        }
-
-
-@dataclass(eq=False)
-class ArrayRechunkState(ShuffleState):
-    type: ClassVar[ShuffleType] = ShuffleType.ARRAY_RECHUNK
-    worker_for: dict[NDIndex, str]
-    old: ChunkedAxes
-    new: ChunkedAxes
-
-    def to_msg(self) -> dict[str, Any]:
-        return {
-            "status": "OK",
-            "type": ArrayRechunkState.type,
-            "run_id": self.run_id,
-            "worker_for": self.worker_for,
-            "old": self.old,
-            "new": self.new,
-            "output_workers": self.output_workers,
-        }
-
-
 class ShuffleSchedulerPlugin(SchedulerPlugin):
     """
     Shuffle plugin for the scheduler
@@ -102,10 +46,11 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
     """
 
     scheduler: Scheduler
-    active_shuffles: dict[ShuffleId, ShuffleState]
+    active_shuffles: dict[ShuffleId, SchedulerShuffleState]
     heartbeats: defaultdict[ShuffleId, dict]
-    _shuffles: defaultdict[ShuffleId, set[ShuffleState]]
-    _archived_by_stimulus: defaultdict[str, set[ShuffleState]]
+    _shuffles: defaultdict[ShuffleId, set[SchedulerShuffleState]]
+    _archived_by_stimulus: defaultdict[str, set[SchedulerShuffleState]]
+    _shift_counter: itertools.count[int]
 
     def __init__(self, scheduler: Scheduler):
         self.scheduler = scheduler
@@ -122,26 +67,38 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
         self.scheduler.add_plugin(self, name="shuffle")
         self._shuffles = defaultdict(set)
         self._archived_by_stimulus = defaultdict(set)
+        self._shift_counter = itertools.count()
 
     async def start(self, scheduler: Scheduler) -> None:
         worker_plugin = ShuffleWorkerPlugin()
         await self.scheduler.register_worker_plugin(
-            None, dumps(worker_plugin), name="shuffle"
+            None, dumps(worker_plugin), name="shuffle", idempotent=False
         )
 
     def shuffle_ids(self) -> set[ShuffleId]:
         return set(self.active_shuffles)
 
-    async def barrier(self, id: ShuffleId, run_id: int) -> None:
+    async def barrier(self, id: ShuffleId, run_id: int, consistent: bool) -> None:
         shuffle = self.active_shuffles[id]
-        assert shuffle.run_id == run_id, f"{run_id=} does not match {shuffle}"
+        if shuffle.run_id != run_id:
+            raise ValueError(f"{run_id=} does not match {shuffle}")
+        if not consistent:
+            logger.warning(
+                "Shuffle %s restarted due to data inconsistency during barrier",
+                shuffle.id,
+            )
+            return self._restart_shuffle(
+                shuffle.id,
+                self.scheduler,
+                stimulus_id=f"p2p-barrier-inconsistent-{time()}",
+            )
         msg = {"op": "shuffle_inputs_done", "shuffle_id": id, "run_id": run_id}
         await self.scheduler.broadcast(
             msg=msg,
             workers=list(shuffle.participating_workers),
         )
 
-    def restrict_task(self, id: ShuffleId, run_id: int, key: str, worker: str) -> dict:
+    def restrict_task(self, id: ShuffleId, run_id: int, key: Key, worker: str) -> dict:
         shuffle = self.active_shuffles[id]
         if shuffle.run_id > run_id:
             return {
@@ -162,7 +119,7 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
             if shuffle_id in self.shuffle_ids():
                 self.heartbeats[shuffle_id][ws.address].update(d)
 
-    def get(self, id: ShuffleId, worker: str) -> dict[str, Any]:
+    def get(self, id: ShuffleId, worker: str) -> ToPickle[ShuffleRunSpec]:
         if worker not in self.scheduler.workers:
             # This should never happen
             raise RuntimeError(
@@ -170,36 +127,41 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
             )  # pragma: nocover
         state = self.active_shuffles[id]
         state.participating_workers.add(worker)
-        return state.to_msg()
+        return ToPickle(state.run_spec)
 
     def get_or_create(
         self,
-        id: ShuffleId,
-        key: str,
-        type: str,
+        # FIXME: This should never be ToPickle[ShuffleSpec]
+        spec: ShuffleSpec | ToPickle[ShuffleSpec],
+        key: Key,
         worker: str,
-        spec: dict[str, Any],
-    ) -> dict:
+    ) -> ToPickle[ShuffleRunSpec]:
+        # FIXME: Sometimes, this doesn't actually get pickled
+        if isinstance(spec, ToPickle):
+            spec = spec.data
         try:
-            return self.get(id, worker)
+            return self.get(spec.id, worker)
         except KeyError:
             # FIXME: The current implementation relies on the barrier task to be
             # known by its name. If the name has been mangled, we cannot guarantee
             # that the shuffle works as intended and should fail instead.
-            self._raise_if_barrier_unknown(id)
+            self._raise_if_barrier_unknown(spec.id)
             self._raise_if_task_not_processing(key)
-
-            state: ShuffleState
-            if type == ShuffleType.DATAFRAME:
-                state = self._create_dataframe_shuffle_state(id, spec)
-            elif type == ShuffleType.ARRAY_RECHUNK:
-                state = self._create_array_rechunk_state(id, spec)
-            else:  # pragma: no cover
-                raise TypeError(type)
-            self.active_shuffles[id] = state
-            self._shuffles[id].add(state)
+            worker_for = self._calculate_worker_for(spec)
+            self._ensure_output_tasks_are_non_rootish(spec)
+            state = spec.create_new_run(
+                worker_for=worker_for, span_id=self.scheduler.tasks[key].group.span_id
+            )
+            self.active_shuffles[spec.id] = state
+            self._shuffles[spec.id].add(state)
             state.participating_workers.add(worker)
-            return state.to_msg()
+            logger.warning(
+                "Shuffle %s initialized by task %r executed on worker %s",
+                spec.id,
+                key,
+                worker,
+            )
+            return ToPickle(state.run_spec)
 
     def _raise_if_barrier_unknown(self, id: ShuffleId) -> None:
         key = barrier_key(id)
@@ -212,43 +174,19 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
                 "into this by leaving a comment at distributed#7816."
             )
 
-    def _raise_if_task_not_processing(self, key: str) -> None:
+    def _raise_if_task_not_processing(self, key: Key) -> None:
         task = self.scheduler.tasks[key]
         if task.state != "processing":
             raise RuntimeError(f"Expected {task} to be processing, is {task.state}.")
 
-    def _create_dataframe_shuffle_state(
-        self, id: ShuffleId, spec: dict[str, Any]
-    ) -> DataFrameShuffleState:
-        column = spec["column"]
-        npartitions = spec["npartitions"]
-        parts_out = spec["parts_out"]
-        assert column is not None
-        assert npartitions is not None
-        assert parts_out is not None
-
-        pick_worker = partial(get_worker_for_range_sharding, npartitions)
-
-        mapping = self._pin_output_workers(id, parts_out, pick_worker)
-        output_workers = set(mapping.values())
-
-        return DataFrameShuffleState(
-            id=id,
-            run_id=next(ShuffleState._run_id_iterator),
-            worker_for=mapping,
-            column=column,
-            output_workers=output_workers,
-            participating_workers=output_workers.copy(),
-        )
-
-    def _pin_output_workers(
-        self,
-        id: ShuffleId,
-        output_partitions: Iterable[Any],
-        pick: Callable[[Any, Sequence[str]], str],
-    ) -> dict[Any, str]:
+    def _calculate_worker_for(self, spec: ShuffleSpec) -> dict[Any, str]:
         """Pin the outputs of a P2P shuffle to specific workers.
 
+        The P2P implementation of a hash join combines the loading of shuffled output
+        partitions for the left and right side with the actual merge operation into a
+        single output task. As a consequence, we need to make sure that shuffles with
+        shared output tasks align on the output mapping.
+
         Parameters
         ----------
         id: ID of the shuffle to pin
@@ -259,68 +197,109 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
             This function assumes that the barrier task and the output tasks share
             the same worker restrictions.
         """
-        mapping = {}
-        barrier = self.scheduler.tasks[barrier_key(id)]
+        existing: dict[Any, str] = {}
+        shuffle_id = spec.id
+        barrier = self.scheduler.tasks[barrier_key(shuffle_id)]
 
         if barrier.worker_restrictions:
             workers = list(barrier.worker_restrictions)
         else:
             workers = list(self.scheduler.workers)
 
-        for partition in output_partitions:
-            worker = pick(partition, workers)
-            mapping[partition] = worker
+        # Ensure homogeneous cluster utilization when there are multiple small,
+        # independent shuffles going on at the same time, e.g. due to partial rechunking
+        shift_by = next(self._shift_counter) % len(workers)
+        workers = workers[shift_by:] + workers[:shift_by]
 
-        for dt in barrier.dependents:
-            try:
-                partition = dt.annotations["shuffle"]
-            except KeyError:
-                continue
+        # Check if this shuffle shares output tasks with a different shuffle that has
+        # already been initialized and needs to be taken into account when
+        # mapping output partitions to workers.
+        # Naively, you could delete this whole paragraph and just call
+        # spec.pick_worker; it would return two identical sets of results on both calls
+        # of this method... until the set of available workers changes between the two
+        # calls, which would cause misaligned shuffle outputs and a deadlock.
+        seen = {barrier}
+        for dependent in barrier.dependents:
+            for possible_barrier in dependent.dependencies:
+                if possible_barrier in seen:
+                    continue
+                seen.add(possible_barrier)
+                if not (other_barrier_key := id_from_key(possible_barrier.key)):
+                    continue
+                if not (shuffle := self.active_shuffles.get(other_barrier_key)):
+                    continue
+                current_worker_for = shuffle.run_spec.worker_for
+                # This is a fail-safe for future three-ways merges. At the moment there
+                # should only ever be at most one other shuffle that shares output
+                # tasks, so existing will always be empty.
+                if existing:  # pragma: nocover
+                    for shared_key in existing.keys() & current_worker_for.keys():
+                        if existing[shared_key] != current_worker_for[shared_key]:
+                            raise RuntimeError(
+                                f"Failed to initialize shuffle {spec.id} because "
+                                "it cannot align output partition mappings between "
+                                f"existing shuffles {seen}. "
+                                f"Mismatch encountered for output partition {shared_key!r}: "
+                                f"{existing[shared_key]} != {current_worker_for[shared_key]}."
+                            )
+                existing.update(current_worker_for)
 
-            if dt.worker_restrictions:
-                worker = pick(partition, list(dt.worker_restrictions))
-                mapping[partition] = worker
-            else:
-                worker = mapping[partition]
+        worker_for = {}
+        for partition in spec.output_partitions:
+            if (worker := existing.get(partition, None)) is None:
+                worker = spec.pick_worker(partition, workers)
+            worker_for[partition] = worker
+        return worker_for
 
-            self._set_restriction(dt, worker)
+    def _ensure_output_tasks_are_non_rootish(self, spec: ShuffleSpec) -> None:
+        """Output tasks are created without worker restrictions and run once with the
+        only purpose of setting the worker restriction and then raising Reschedule, and
+        then running again properly on the correct worker. It would be non-trivial to
+        set the worker restriction before they're first run due to potential task
+        fusion.
 
-        return mapping
+        Most times, this lack of initial restrictions would cause output tasks to be
+        labelled as rootish on their first (very fast) run, which in turn would break
+        the design assumption that the worker-side queue of rootish tasks will last long
+        enough to cover the round-trip to the scheduler to receive more tasks, which in
+        turn would cause a measurable slowdown on the overall runtime of the shuffle
+        operation.
 
-    def _create_array_rechunk_state(
-        self, id: ShuffleId, spec: dict[str, Any]
-    ) -> ArrayRechunkState:
-        old = spec["old"]
-        new = spec["new"]
-        assert old is not None
-        assert new is not None
+        This method ensures that, given M output tasks and N workers, each worker-side
+        queue is pre-loaded with M/N output tasks which can be flushed very fast as
+        they all raise Reschedule() in quick succession.
 
-        parts_out = product(*(range(len(c)) for c in new))
-        mapping = self._pin_output_workers(id, parts_out, get_worker_for_hash_sharding)
-        output_workers = set(mapping.values())
-
-        return ArrayRechunkState(
-            id=id,
-            run_id=next(ShuffleState._run_id_iterator),
-            worker_for=mapping,
-            output_workers=output_workers,
-            old=old,
-            new=new,
-            participating_workers=output_workers.copy(),
-        )
+        See Also
+        --------
+        ShuffleRun._ensure_output_worker
+        """
+        barrier = self.scheduler.tasks[barrier_key(spec.id)]
+        for dependent in barrier.dependents:
+            dependent._rootish = False
 
+    @log_errors()
     def _set_restriction(self, ts: TaskState, worker: str) -> None:
-        if "shuffle_original_restrictions" in ts.annotations:
+        if ts.annotations and "shuffle_original_restrictions" in ts.annotations:
             # This may occur if multiple barriers share the same output task,
             # e.g. in a hash join.
             return
-        ts.annotations["shuffle_original_restrictions"] = ts.worker_restrictions.copy()
+        if ts.annotations is None:
+            ts.annotations = dict()
+        ts.annotations["shuffle_original_restrictions"] = (
+            ts.worker_restrictions.copy()
+            if ts.worker_restrictions is not None
+            else None
+        )
         self.scheduler.set_restrictions({ts.key: {worker}})
 
+    @log_errors()
     def _unset_restriction(self, ts: TaskState) -> None:
         # shuffle_original_restrictions is only set if the task was first scheduled
         # on the wrong worker
-        if "shuffle_original_restrictions" not in ts.annotations:
+        if (
+            ts.annotations is None
+            or "shuffle_original_restrictions" not in ts.annotations
+        ):
             return
         original_restrictions = ts.annotations.pop("shuffle_original_restrictions")
         self.scheduler.set_restrictions({ts.key: original_restrictions})
@@ -360,6 +339,7 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
         recs = self._restart_recommendations(id)
         self.scheduler.transitions(recs, stimulus_id=stimulus_id)
         self.scheduler.stimulus_queue_slots_maybe_opened(stimulus_id=stimulus_id)
+        logger.warning("Shuffle %s restarted due to stimulus '%s", id, stimulus_id)
 
     def remove_worker(
         self, scheduler: Scheduler, worker: str, *, stimulus_id: str, **kwargs: Any
@@ -380,6 +360,12 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
         for shuffle_id, shuffle in self.active_shuffles.copy().items():
             if worker not in shuffle.participating_workers:
                 continue
+            logger.debug(
+                "Worker %s removed during active shuffle %s due to stimulus '%s'",
+                worker,
+                shuffle_id,
+                stimulus_id,
+            )
             exception = RuntimeError(f"Worker {worker} left during active {shuffle}")
             self._fail_on_workers(shuffle, str(exception))
             self._clean_on_scheduler(shuffle_id, stimulus_id)
@@ -389,7 +375,7 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
 
     def transition(
         self,
-        key: str,
+        key: Key,
         start: TaskStateState,
         finish: TaskStateState,
         *args: Any,
@@ -399,13 +385,21 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
         """Clean up scheduler and worker state once a shuffle becomes inactive."""
         if finish not in ("released", "forgotten"):
             return
-        if not key.startswith("shuffle-barrier-"):
-            return
         shuffle_id = id_from_key(key)
+        if not shuffle_id:
+            return
 
         if shuffle := self.active_shuffles.get(shuffle_id):
             self._fail_on_workers(shuffle, message=f"{shuffle} forgotten")
             self._clean_on_scheduler(shuffle_id, stimulus_id=stimulus_id)
+            logger.debug(
+                "Shuffle %s forgotten because task %r transitioned to %s due to "
+                "stimulus '%s'",
+                shuffle_id,
+                key,
+                finish,
+                stimulus_id,
+            )
 
         if finish == "forgotten":
             shuffles = self._shuffles.pop(shuffle_id, set())
@@ -416,7 +410,7 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
                     if not archived:
                         del self._archived_by_stimulus[shuffle._archived_by]
 
-    def _fail_on_workers(self, shuffle: ShuffleState, message: str) -> None:
+    def _fail_on_workers(self, shuffle: SchedulerShuffleState, message: str) -> None:
         worker_msgs = {
             worker: [
                 {
@@ -430,9 +424,10 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
         }
         self.scheduler.send_all({}, worker_msgs)
 
-    def _clean_on_scheduler(self, id: ShuffleId, stimulus_id: str | None) -> None:
+    def _clean_on_scheduler(self, id: ShuffleId, stimulus_id: str) -> None:
         shuffle = self.active_shuffles.pop(id)
-        if not shuffle._archived_by and stimulus_id:
+        logger.warning("Shuffle %s deactivated due to stimulus '%s'", id, stimulus_id)
+        if not shuffle._archived_by:
             shuffle._archived_by = stimulus_id
             self._archived_by_stimulus[stimulus_id].add(shuffle)
 
@@ -448,19 +443,3 @@ class ShuffleSchedulerPlugin(SchedulerPlugin):
         self.heartbeats.clear()
         self._shuffles.clear()
         self._archived_by_stimulus.clear()
-
-
-def get_worker_for_range_sharding(
-    npartitions: int, output_partition: int, workers: Sequence[str]
-) -> str:
-    """Get address of target worker for this output partition using range sharding"""
-    i = len(workers) * output_partition // npartitions
-    return workers[i]
-
-
-def get_worker_for_hash_sharding(
-    output_partition: NDIndex, workers: Sequence[str]
-) -> str:
-    """Get address of target worker for this output partition using hash sharding"""
-    i = hash(output_partition) % len(workers)
-    return workers[i]
